/*
 *	Author: Jan Dolezel, http://www.eithel.net
 *	Last version: http://www.eithel.net/download/ui/spiral
 *	Last modified: 2005-11-30
 *
 *	License Information:
 *	-------------------------------------------------------------------------
 *	Copyright (C) 2005 Jan Dolezel, Miroslav Skrbek
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2 of the License, or (at your
 *	option) any later version.
 *
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *	General Public License for more details.
*/

package tablet;

import java.io.*;

/**
 * <p>Class stores information about tablet packet/point.</p>
 *
 * <p>Copyright: Copyright (c) 2005 Miroslav Skrbek, Jan Dolezel</p>
 * <p>Company: VUT FEL, Prague</p>
 *
 * @author Miroslav Skrbek
 * @author Jan Dolezel
 * @version 1.1.3
 */
public class TabletPoint {
  /**
   * pen is touching a surface
   */
  public static final int BUTTON_TOUCH = 0x0001;
  /**
   * 1st button is pressed
   */
  public static final int BUTTON_1 = 0x0001;
  /**
   * 2nd button is pressed
   */
  public static final int BUTTON_2 = 0x0002;
  /**
   * 3rd button is pressed
   */
  public static final int BUTTON_3 = 0x0004;

  /**
   * x coordinate of point
   */
  public static final int PK_X = 0;
  /**
   * y coordinate of point
   */
  public static final int PK_Y = 1;
  /**
   * pressed buttons
   */
  public static final int PK_BUTTONS = 2;
  /**
   * pen's slope
   */
  public static final int PK_AZIMUTH = 3;
  /**
   * altitude of pen above surface
   */
  public static final int PK_ALTITUDE = 4;
  /**
   * pen's pressure
   */
  public static final int PK_NORMAL_PRESSURE = 5;
  /**
   * packet's creation time
   */
  public static final int PK_TIME = 6;
  /**
   * packet's serial number
   */
  public static final int PK_SERIAL_NUMBER = 7;

  /**
   * packet with information about point
   */
  private int[] packet = new int[8];

  /**
   * creates new point from tablet packet
   * <p>Format of packet:<br>
   * 0 - x coordinate<br>
   * 1 - y coordinate<br>
   * 2 - buttons<br>
   * 3 - azimuth<br>
   * 4 - altitude<br>
   * 5 - pressure<br>
   * 6 - time<br>
   * 7 - serial number</p>
   */
  public TabletPoint(int[] packet) {
    this.packet = new int[packet.length];
    System.arraycopy(packet, 0, this.packet, 0, packet.length);
  }

  /**
   * creates new point from byte array when loading from file
   * <p>Format of byte array:<br>
   * 3 bytes - x coordinate<br>
   * 3 bytes - y coordinate<br>
   * 4 bytes - time<br>
   * 2 bytes - azimuth<br>
   * 2 bytes - altitude<br>
   * 2 bytes - pressure<br>
   * 1 byte  - buttons<br>
   * 3 bytes - serial number</p>
   * @param array 20 bytes of information
   */
  public TabletPoint(byte array[]) {
    packet[PK_X] = this.toInt(array, 3, 0);
    packet[PK_Y] = this.toInt(array, 3, 3);
    packet[PK_TIME] = this.toInt(array, 4, 6);
    packet[PK_AZIMUTH] = this.toInt(array, 2, 10);
    packet[PK_ALTITUDE] = this.toInt(array, 2, 12);
    packet[PK_NORMAL_PRESSURE] = this.toInt(array, 2, 14);
    packet[PK_BUTTONS] = this.toInt(array, 1, 16);
    packet[PK_SERIAL_NUMBER] = this.toInt(array, 3, 17);
  }

  /**
   * returns x coordinate of point
   */
  public int getX() {
    return packet[TabletPoint.PK_X];
  }

  /**
   * returns y coordinate of point
   */
  public int getY() {
    return packet[TabletPoint.PK_Y];
  }

  /**
   * returns pressed buttons
   */
  public int getButtons() {
    return packet[TabletPoint.PK_BUTTONS];
  }

  /**
   * returns slope of pen
   */
  public int getAzimuth() {
    return packet[TabletPoint.PK_AZIMUTH];
  }

  /**
   * returns altitude of pen above surface
   */
  public int getAltitude() {
    return packet[TabletPoint.PK_ALTITUDE];
  }

  /**
   * returns pen's pressure
   */
  public int getNormalPressure() {
    return packet[TabletPoint.PK_NORMAL_PRESSURE];
  }

  /**
   * returns time of packet's creation
   */
  public long getTime() {
    return ( (long) packet[TabletPoint.PK_TIME]) & 0xFFFFFFFF;
  }

  /**
   * returns serial number of packet
   */
  public int getSerialNumber() {
    return packet[TabletPoint.PK_SERIAL_NUMBER];
  }

  /**
   * tests whether is specified button pressed
   */
  public boolean isButtonPressed(int button) {
    return (packet[PK_BUTTONS] & button) != 0;
  }

  public boolean isTouched() {
    return this.isButtonPressed(TabletPoint.BUTTON_TOUCH);
  }

  /**
   * Header with information about text format of TabletPoint
   */
  public static String header() {
    return "#         X          Y       Time    Azimuth   Altitude   Pressure" +
        "    Buttons  SerialNum";
  }

  /**
   * text representation of TabletPoint
   * @return 88 characters
   */
  public String toString() {
    return
        String.format(" %10d %10d %10d %10d %10d %10d %10d %10d",
                      packet[PK_X],
                      packet[PK_Y],
                      packet[PK_TIME],
                      packet[PK_AZIMUTH],
                      packet[PK_ALTITUDE],
                      packet[PK_NORMAL_PRESSURE],
                      packet[PK_BUTTONS],
                      packet[PK_SERIAL_NUMBER]
        );
  }

  /**
   * writes 20 bytes of information to output stream
   * @throws IOException
   */
  public void print(OutputStream out) throws IOException {
    byte pom[] = this.toByte(packet[PK_X], 3);
    out.write(pom, 0, pom.length);
    pom = this.toByte(packet[PK_Y], 3);
    out.write(pom, 0, pom.length);
    pom = this.toByte(packet[PK_TIME], 4);
    out.write(pom, 0, pom.length);
    pom = this.toByte(packet[PK_AZIMUTH], 2);
    out.write(pom, 0, pom.length);
    pom = this.toByte(packet[PK_ALTITUDE], 2);
    out.write(pom, 0, pom.length);
    pom = this.toByte(packet[PK_NORMAL_PRESSURE], 2);
    out.write(pom, 0, pom.length);
    pom = this.toByte(packet[PK_BUTTONS], 1);
    out.write(pom, 0, pom.length);
    pom = this.toByte(packet[PK_SERIAL_NUMBER], 3);
    out.write(pom, 0, pom.length);
  }

  /**
   * writes text representation of point to print stream
   * @throws IOException
   */
  public void print(PrintStream out) throws IOException {
    out.printf("%s%n", this.toString());
  }

  /**
   * converts int array of bytes
   * @param num positive integer to convert
   * @param bytes array length
   * @return byte[] coverted byte array
   */
  private byte[] toByte(int num, int bytes) {
    byte res[] = new byte[bytes];
    for(int i = 0; i < bytes; i++) {
      res[i] = (byte) (num % 256);
      num /= 256;
    }
    return res;
  }

  /**
   * converts part of byte array to positive integer
   * @param array byte array to convert
   * @param bytes number of bytes to convert
   * @param position start position
   * @return int converted integer
   */
  private int toInt(byte[] array, int bytes, int position) {
    int res = 0, pow = 1;
    for(int i = 0; i < bytes; i++) {
      int p = array[position + i];
      p = ( p < 0 ) ? 256+p: p;
      res += p * pow;
      pow *= 256;
    }
    return res;
  }
}
